import os

from jamp.jam_syntax import parse
from jamp.compile import compile
from jamp.executors import run
from jamp.classes import State
from jamp.jam_builtins import Builtins


def expect_output(expected):
    cp = Builtins.output
    print(cp)
    Builtins.clear_output()
    assert cp == expected


def test_simple_rule_with_assigns():
    rule_t = """
    a = 10 ;

    rule A
    {
        local a = 3 ;
        b = 2 ;
        a = 4 ;
        local c  ;

        Echo "inside a" $(a) ;
        Echo "inside b" $(b) ;
    }

    actions A {
        echo "actions" ;
    }

    Echo "outside a" $(a) ;
    Echo "outside b" $(b) ;

    A ;

    Echo "outside a" $(a) ;
    Echo "outside b" $(b) ;
    Echo "test" "test1" "test2" ;
    """

    state = State()
    ast = parse(rule_t)
    cmds = compile(state, ast)
    run(state, cmds)

    output = """outside a 10
outside b
inside a 4
inside b 2
outside a 10
outside b 2
test test1 test2
"""

    expect_output(output)


def test_rule_targets():
    rule_t = """
    rule A
    {
        Echo $(<) ;
        Echo $(>) ;
        Echo $(1) ;
        Echo $(2) ;
    }

    Echo $(<) ;
    A "one" : two ;
    """

    state = State()
    run(state, state.parse_and_compile(rule_t))
    output = """
one
two
one
two
"""
    expect_output(output)


def test_simple_if():
    rules = """
    a = 4 ;
    b = 4 ;
    c = 4 ;

    if $(a) = $(b)
    {
        local a = "check" ;
        local b = "ok" ;
        Echo $(a) $(b) ;
    }

    Echo $(a) $(b) : $(c) ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))

    output = """check ok
4 4 4
"""

    expect_output(output)


def test_simple_else():
    rules = """
    a = 4 ;
    b = 5 ;

    if $(a) = $(b)
    {
        Echo "equal" ;
    }
    else
    {
        Echo "not equal" ;
    }

    if $(a) = $(b)
    {
        Echo "equal" ;
    }
    else
        Echo "not equal 2" ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))

    output = "not equal\nnot equal 2\n"
    expect_output(output)


def test_comparisons_and_not():
    rules = """
    a = 4 ;
    b = 4 ;
    c = 4 4 ;
    d = 4 4 ;
    e = 5 4 ;

    if ! ( $(a) != $(b) )
    {
        Echo "check 1 ok" ;
    }

    if $(c) > $(b)
    {
        Echo "check 2 ok" ;
    }

    if $(c) >= $(d)
    {
        Echo "check 3 ok" ;
    }

    if $(c) < $(e)
    {
        Echo "check 4 ok" ;
    }

    if $(c) <= $(d)
    {
        Echo "check 5 ok" ;
    }

    if $(c) > $(e) || $(a) = $(b)
    {
        Echo "check 6 ok" ;
    }

    if $(c) > $(e) && $(a) = $(b)
    {
        Echo "check 7 fail" ;
    }

    if ! ( $(c) > $(e) && $(a) = $(b) )
    {
        Echo "check 8 ok" ;
    }

    if $(a) in $(e)
    {
        Echo "check 9 ok" ;
    }

    if $(e) in $(a)
    {
        Echo "check 10 not ok" ;
    }
    """

    state = State()
    run(state, state.parse_and_compile(rules))

    output = """check 1 ok
check 2 ok
check 3 ok
check 4 ok
check 5 ok
check 6 ok
check 8 ok
check 9 ok
"""

    expect_output(output)


def test_while():
    rules = """
    a = 4 ;

    while $(a)
    {
        Echo $(a) ;

        if $(a) = 6
        {
            a = ;
        }

        if $(a) = 5
        {
            a = 6 ;
        }

        if $(a) = 4
        {
            a = 5 ;
        }
    }
    """

    state = State()
    run(state, state.parse_and_compile(rules))

    output = """4\n5\n6\n"""
    expect_output(output)


def test_while_break():
    rules = """
    a = 4 ;

    while $(a)
    {
        Echo $(a) ;

        if $(a) = 6
        {
            break ;
        }

        if $(a) = 5
        {
            a = 6 ;
        }

        if $(a) = 4
        {
            a = 5 ;
        }
    }
    """

    state = State()
    run(state, state.parse_and_compile(rules))

    output = """4\n5\n6\n"""
    expect_output(output)


def test_while_continue():
    rules = """
    a = 4 ;

    while $(a)
    {
        Echo $(a) ;

        if $(a) = 6
        {
            break ;
        }

        if $(a) = 5
        {
            a = 6 ;
            continue ;
            a = 11 ;
        }

        if $(a) = 4
        {
            a = 5 ;
            continue ;
        }

        a = 10 ;
    }
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    expect_output("4\n5\n6\n")


def test_for():
    rules = """
    items = 1 2 3 4 ;

    for item in $(items)
    {
        local a ;

        if ! $(a)
        {
            a = 10 ;
        }

        Echo $(a) ;
        a = $(item) ;
        Echo $(a) ;
    }

    Echo "item" $(item) ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    output = """10
1
10
2
10
3
10
4
item 4
"""
    expect_output(output)


def test_switch():
    rules = """
    a = "test1" ;
    switch $(a)
    {
        case ?est2 :
            Echo "test2" ;
        case ?est1 :
            Echo "test1" ;
        case ?est3 :
            Echo "test3" ;
    }
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = """test1\n"""
    expect_output(output)


def test_params():
    rules = """
    rule check
    {
        Echo $(<) ;
        Echo $(>) ;
        Echo $(<[2]) $(<[4]) ;
    }

    check "1" "2" "3" "4" "5" : "11" "22" "33" ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "1 2 3 4 5\n11 22 33\n2 4\n"
    expect_output(output)


def test_params_only_left():
    rules = """
    rule check
    {
        Echo $(<[2]) $(<[4]) ;
    }

    check "1" "2" "3" "4" "5"  ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "2 4\n"
    expect_output(output)


def test_include():
    rules = """
    rule Test
    {
        Echo "1" ;
    }

    Test1 "aaa" ;

    include tests/data/include_test.jam ;

    Test ;
    Test1 "aaa" ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    output = "jamp: unknown rule Test1\n1\n2\n"
    expect_output(output)


def test_assign_types():
    rules = """
    a = 1 ;
    local b ;
    local c ;
    a ?= 2 ;
    Echo $(a) ;
    a += 3 ;
    Echo $(a) ;
    b += 4 ;
    Echo $(b) ;
    c += ;
    Echo $(c) ;
    a += 5 ;
    Echo $(a) ;
    d ?= 6 7 ;
    Echo $(d) ;
    e ?= 8 ;
    Echo $(e) ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    output = "1\n1 3\n4\n\n1 3 5\n6 7\n8\n"
    expect_output(output)


def test_assign_types_on_target():
    rules = """
    a on test.c = 1 ;
    a on test.c ?= 2 ;
    b on test.c = ;
    c on test.c = ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    target = state.get_target("test.c")
    assert target.vars["a"] == ["1"]

    rules = """
    a on test.c += 3 ;
    b on test.c += 1 ;
    c on test.c += 1 5 ;
    """

    run(state, state.parse_and_compile(rules))
    assert target.vars["a"] == ["1", "3"]
    assert target.vars["b"] == ["1"]
    assert target.vars["c"] == ["1", "5"]


def test_on_target():
    rules = """
    a on test.c = "val1" ;
    b on test2.c = "val3" ;
    c = "val2" ;

    on test.c Echo $(a) $(b) $(c) ;
    on test2.c
    {
        Echo $(a) $(b) $(c) ;
    }
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "val1 val2\nval3 val2\n"
    expect_output(output)


def test_func():
    rules = """
    rule some_func
    {
        a = 1 2 ;
        return $(1) $(a) ;
    }

    res = [ some_func "one" ] ;
    Echo $(res) ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "one 1 2\n"
    expect_output(output)


def test_rule_cycle_return():
    rules = """
    rule one
    {
        Echo $(<) ;
        a = $(<) $(>) ;

        for i in $(a)
        {
            Echo $(i) ;
        }

        return $(a) $(i) ;
    }

    res = [ one 1 2 : 3 4 : 5 6 ] ;
    echo $(res) ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "1 2\n1\n2\n3\n4\n1 2 3 4 4\n"
    expect_output(output)


def test_glob_rule():
    rules = """
    res = [ glob . : *.py *.ini ] ;
    echo $(res) ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    expected = set(f".{os.sep}pytest.ini .{os.sep}conftest.py\n".strip().split(" "))
    cp = set(Builtins.output.strip().split(" "))
    Builtins.clear_output()
    assert cp == expected


def test_match_rule():
    rules = r"""
    res = [ match \\d+ ab : 23adf abba adad ] ;
    echo $(res) ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "23 ab\n"
    expect_output(output)


def test_expanded_rule_and_empty_result():
    rules = r"""
    rule One
    {
        return ;
    }
    a = ;
    res = [ $(a) "one" "two" ] ;
    Echo $(res) 1 ;
    res = [ One "one" "two" ] ;
    Echo $(res) 2 ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "1\n2\n"
    expect_output(output)


def test_expanded_rule():
    rules = r"""
    rule one
    {
        return 1 2 ;
    }
    rule two
    {
        return  ;
    }
    rule three
    {
        return 3 ;
    }

    a = one two three ;
    res = [ $(a) "nothing" ] ;
    Echo $(res) ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    assert state.vars.get("res") == ["1", "2", "3"]
    output = "1 2 3\n"
    expect_output(output)


def test_expanded_with_empty():
    rules = r"""
    local a = a1 a2 a3 ;
    local b = ;
    local c = c1 ;
    local d = d1 ;

    Echo "1" "/define=( $(a) )" ;
    Echo "2" "/define=( $(b) )" ;
    Echo "3" "/define=( $(a) $(b) )" ;
    Echo "4" "/define=( $(b) $(a) )" ;
    Echo "5" $(c)$(b) ;
    Echo "6" $(c)$(d) ;
    Echo "7" $(c)$(b)$(d) ;
    Echo "8" $(c)$(d)$(b) ;
    Echo "9" "$(c)$(b)$(d)" ;
    Echo "10" "$(c)$(d)$(b)" ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = """1 /define=( a1 ) /define=( a2 ) /define=( a3 )
2
3
4
5
6 c1d1
7
8
9
10
"""
    expect_output(output)


def test_assign_empty():
    rules = r"""
    a = "" ;
    b = ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    assert state.vars.get("a") == [""]
    assert state.vars.get("b") == []


def test_quoted_empty():
    rules = r"""
    rule Test { return "/define=( $(<:J=,) )" ; }

    local a = ;
    Echo [ Test $(a) ] ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    output = "\n"
    expect_output(output)


def test_quoted_non_empty():
    rules = r"""
    rule Test { return "/define=( $(<:J=,) )" ; }

    local a = "inc1:" "inc2:" ;
    Echo [ Test $(a) ] ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    output = "/define=( inc1:,inc2: )\n"
    expect_output(output)


def test_on_return():
    rules = """
    AR = ar ;

    actions Archive
    {
        $(AR)
    }

    Archive test.a : test.obj ;
    AR on test.a = "ar ru" ;
    Echo [ on test.a return $(AR) "some" ] ;
    Echo "ok" ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "ar ru some\nok\n"
    expect_output(output)


def test_empty_1():
    rules = """a = ;
b = "" ;
c = "" "" ;

Echo $(a)1 ;
Echo $(b)2 ;
Echo $(c)3 ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "\n2\n3 3\n"
    expect_output(output)


def test_empty_2():
    rules = """a = ;
b = "" ;
c = "" "" ;
d = $(c) "" ;

Echo $(a)1 ;
Echo $(b)2 ;
Echo $(c)3 ;
Echo $(d)4 ;
    """
    state = State()
    run(state, state.parse_and_compile(rules))
    output = "\n2\n3 3\n4 4 4\n"
    expect_output(output)



import sys
import os
import re
import subprocess as sp

from typing import Union
from pathlib import Path

from jamp.classes import State, Target, Exec
from jamp.jam_syntax import Arg, Node
from jamp.expand import var_expand, lol_get
from jamp.pattern import match


def output(text, end="\n"):
    if not hasattr(sys, "_called_from_test"):
        print(text, end=end)
    else:
        Builtins.output += text + end


class Builtins:
    """Singleton with builtin rules"""

    output = ""

    def __init__(self):
        self.dir_counter = 0

    @classmethod
    def clear_output(cls):
        cls.output = ""

    def glob(self, state: State, lol: list):
        from jamp.executors import Result

        dirs = lol_get(lol, 0)
        patterns = lol_get(lol, 1)

        if len(patterns) == 0 or len(dirs) == 0:
            return []

        res = []
        for d in dirs:
            p = Path(d)
            if not p.is_dir():
                continue

            for f in p.iterdir():
                for m in patterns:
                    if match(m, f.name) == 0:
                        res.append(os.path.join(d, f.name))
                        break

        return Result(res)

    def mkdir(self, state: State, paths_arg: list):
        from jamp.paths import Pathname

        paths = expand(state, paths_arg)

        dirs_target = state.targets.get("dirs")
        if dirs_target is None:
            print("fatal error, dirs target not found")
            return

        dirs_target.is_dirs_target = True

        for path in paths:
            p = Pathname()
            p.parse(paths[0])
            p.grist = ""

            dry = p.build()
            if dry == "." or dry == "..":
                continue

            if dirs_target.collected_dirs is None:
                dirs_target.collected_dirs = set()

            dirs_target.collected_dirs.add(dry)

    def pathexists(self, state, paths_arg):
        from jamp.executors import Result
        from jamp.paths import Pathname

        paths = expand(state, paths_arg)
        res = []
        if len(paths):
            p = Pathname()
            p.parse(paths[0])
            p.grist = ""
            if os.path.exists(p.build()):
                res = ["1"]

        return Result(res)

    def match(self, state: State, lol: list):
        from jamp.executors import Result

        strings = lol_get(lol, 1)
        patterns = lol_get(lol, 0)

        if len(patterns) == 0 or len(strings) == 0:
            return []

        if not hasattr(self, "match_complained"):
            print("jamp: Match rule works by Python regular expression rules")
            self.match_complained = True

        res = []
        for s in strings:
            for p in patterns:
                matches = re.findall(p, s)
                if not matches:
                    continue

                if isinstance(matches[0], str):
                    res += [m for m in matches if m]
                elif isinstance(matches[0], tuple):
                    for m in matches:
                        for innertup in m:
                            val = "".join(innertup)
                            if val:
                                res.append(val)

        return Result(res)

    def depends(self, state: State, lol: list, includes=False):
        """
        Depends() - DEPENDS/INCLUDES rule

        The DEPENDS builtin rule appends each of the listed sources on the
        dependency list of each of the listed targets.  It binds both the
        targets and sources as TARGETs.

        TODO: use includes
        """

        targets = lol_get(lol, 0)
        sources = lol_get(lol, 1)

        if sources == ["."]:
            return

        if targets == sources:
            return

        for target_name in targets:
            target = Target.bind(state, target_name)

            if includes:
                target.add_includes(state, sources)
            else:
                target.add_depends(state, sources)

    def includes(self, state: State, args: list):
        return self.depends(state, args, includes=True)

    def always(self, state: State, targets: list):
        targets = expand(state, targets)
        for target_name in targets:
            target = Target.bind(state, target_name)
            state.always_build.add(target)

    def notfile(self, state: State, targets: list):
        """Set the targets as phony"""

        targets = expand(state, targets)
        for target_name in targets:
            target = Target.bind(state, target_name)
            target.notfile = True

    def generated(self, state: State, targets: list):
        """Force targets use generator in their rules"""

        targets = expand(state, targets)
        for target_name in targets:
            target = Target.bind(state, target_name)
            target.generated = True

    def restat(self, state: State, targets: list):
        """Force targets use restat in their rules"""

        targets = expand(state, targets)
        for target_name in targets:
            target = Target.bind(state, target_name)
            target.restat = True

    def temporary(self, state: State, targets: list):
        targets = expand(state, targets)
        for target_name in targets:
            target = Target.bind(state, target_name)
            target.temporary = True

    def leaves(self, state: State, targets: list):
        if hasattr(self, "leaves_complained"):
            return

        print('jamp: "leaves" rule is ignored')
        self.leaves_complained = True

    def nocare(self, state: State, targets: list):
        targets = expand(state, targets)
        for target_name in targets:
            target = Target.bind(state, target_name)
            target.nocare = True

    def noupdate(self, state: State, targets: list):
        targets = expand(state, targets)
        for target_name in targets:
            target = Target.bind(state, target_name)
            target.noupdate = True

    def _print(self, val: Union[str, list], first: bool = True):
        """unflatten before printing"""
        if isinstance(val, str):
            if not first:
                output(" ", end="")

            output(val, end="")
            first = False
        else:
            for item in val:
                first = self._print(item, first)

        return first

    def echo(self, state: State, params: list):
        self._print(params)
        output("")

    def exit(self, state, args):
        self.echo(state, args)
        exit(1)

    def pdb(self, state, args):
        from jamp.executors import FLOW_DEBUG

        return FLOW_DEBUG

    def command(self, state: State, args):
        from jamp.executors import Result

        cmd = " ".join(expand(state, args))
        output = ""
        try:
            output = sp.check_output(cmd, shell=True)
            output = output.decode("utf8").strip()
        except sp.CalledProcessError:
            if state.verbose:
                print(f"jamp: command returned non-zero status:\n{cmd}\n{output}")
            elif not hasattr(self, "nonzero_complained"):
                print(
                    "jamp: some of Command rules returned "
                    "non-zero status (use --verbose for more info)"
                )
                self.nonzero_complained = True

        return Result([output])


def flatten(res: list):
    # try to remove one level
    if isinstance(res, list) and len(res) == 1:
        res = res[0]

    return res


def iter_var(var, skip_empty=True):
    if isinstance(var, str):
        yield var
    elif isinstance(var, list):
        for item in var:
            yield item
    else:
        raise Exception(f"unexpected var for iteration: {var}")


def validate(var):
    if not isinstance(var, list):
        raise Exception(f"validation: expected list, got: {var}")
    if len(var) > 0:
        if not isinstance(var[0], str):
            raise Exception(f"validation: expected str on second level, got: {var[0]}")


def validate_lol(var):
    if not isinstance(var, list):
        raise Exception(f"LOL validation: expected list, got: {var}")
    if len(var) > 0:
        if not isinstance(var[0], list):
            raise Exception(
                f"LOL validation: expected list on second level, got: {var[0]}"
            )


def expand(state: State, arg: Union[Arg, tuple, str], skip_empty=True):
    """Make list of strings from some type of an argument"""

    if arg is None or (skip_empty and arg == ""):
        return []
    elif arg == "":
        return [""]
    elif isinstance(arg, str):
        res = var_expand(arg, state.params, state.vars)
    elif isinstance(arg, Exec):
        from jamp.executors import Result

        execval = arg.execute(state)
        if execval is None:
            res = []
        elif isinstance(execval, Result):
            res = execval.val
        else:
            raise Exception(f"expected result, got {execval}")
    elif isinstance(arg, Arg):
        res = expand(state, arg.value, skip_empty=skip_empty)
    elif isinstance(arg, list):
        res = []
        for item in arg:
            val = expand(state, item, skip_empty=skip_empty)
            for v in iter_var(val, skip_empty=skip_empty):
                res.append(v)
    else:
        print(type(arg))
        raise Exception(f"could not expand arg: {arg}")

    validate(res)
    return res


def expand_lol(state: State, arg: tuple):
    """Make string from some type of an argument"""

    if isinstance(arg, tuple) and len(arg) and arg[0] == Node.LOL:
        res = []
        for lol_list in arg[1:]:
            if lol_list:
                res.append(expand(state, lol_list))
    elif isinstance(arg, list):
        res = [[expand(state, item) for item in arg]]
    else:
        raise Exception(f"could not expand LOL: {arg}")

    validate_lol(res)
    return res


def re_expand(state: State, string: str, params: list):
    def exp(m: re.Match[str]) -> str:
        expanded = var_expand(m.string[m.start() : m.end()], params, state.vars)
        return " ".join(expanded)

    return re.sub(r"\$\([^\(]+\)", exp, string)



from os import sep as S

from jamp.expand import var_expand, var_edit_parse, MAGIC_COLON as MC
from jamp.paths import Pathname


def test_nested():
    assert var_expand("$(a)", [], {"a": "1"}) == ["1"]
    assert var_expand("$($(a))", [], {"a": "b", "b": "1"}) == ["1"]
    assert var_expand("$($($(a)))", [], {"a": "b", "b": "c", "c": "1"}) == ["1"]


def test_params():
    lol = [["i", str(i + 1)] for i in range(10)]
    assert var_expand("$(<)", ["a", "b"], None) == "a"
    assert var_expand("$(>)", ["a", "b"], None) == "b"
    assert var_expand("$(1)", ["a", "b", "c"], None) == "a"
    assert var_expand("$(2)", ["a", "b", "c"], None) == "b"
    assert var_expand("$(3)", lol, None) == ["i", "3"]
    assert var_expand("$(4)", lol, None) == ["i", "4"]
    assert var_expand("$(8)", lol, None) == ["i", "8"]
    assert var_expand("$(9)", lol, None) == ["i", "9"]
    assert var_expand("$(10)", ["a", "b"], {"10": "tt"}) == ["tt"]
    assert var_expand("$(a[2-])", [], {"a": ["1", "2", "3", "4"]}) == ["2", "3", "4"]
    assert var_expand("$(a[2])", [], {"a": ["1", "2", "3"]}) == ["2"]
    assert var_expand("$(a[2-2])", [], {"a": ["1", "2", "3", "4"]}) == ["2"]
    assert var_expand("$(a[2-1])", [], {"a": ["1", "2", "3", "4"]}) == []
    assert var_expand("$(a[2-6])", [], {"a": ["1", "2", "3", "4"]}) == ["2", "3", "4"]
    assert var_expand("$(a[2-3])", [], {"a": ["1", "2", "3", "4"]}) == ["2", "3"]
    assert var_expand("$(a[1])", [], {"a": ["1", "2", "3", "4"]}) == ["1"]
    assert var_expand("$(a[4])", [], {"a": ["1", "2", "3", "4"]}) == ["4"]
    assert var_expand("$(a[5])", [], {"a": ["1", "2", "3", "4"]}) == []
    assert var_expand("ad$(a[4])da", [], {"a": ["1", "2", "3", "4"]}) == ["ad4da"]
    assert var_expand("ad$(a)da", [], {"a": ["1", "2"]}) == ["ad1da", "ad2da"]
    assert var_expand("ad$(a)da$(b)gh", [], {"a": ["1", "2"], "b": ["3", "4"]}) == [
        "ad1da3gh",
        "ad1da4gh",
        "ad2da3gh",
        "ad2da4gh",
    ]


def test_empty_val():
    assert var_expand("-$(a)", [], {}) == []
    assert var_expand("-$(a)$(b)", [], {}) == []
    assert var_expand("-$(a)-", [], {}) == []
    assert var_expand("-$(a)$(b)", [], {"b": "1"}) == []
    assert var_expand("$(a)$(b)$(c)", [], {"a": ["1"], "b": [], "c": ["3"]}) == []
    assert var_expand("$(a)$(b)$(c)", [], {"a": ["1"], "b": [""], "c": ["3"]}) == ["13"]


def test_expand2():
    assert var_expand("-$(h)", [], {"h": ["val1", "val2"]}) == ["-val1", "-val2"]
    assert var_expand("-$(h)-", [], {"h": ["val1", "val2"]}) == ["-val1-", "-val2-"]
    assert var_expand("-$($(h))", [], {"h": ["a", "b"], "a": "1", "b": "2"}) == [
        "-1",
        "-2",
    ]
    assert var_expand("-$(a)$($(h))", [], {"h": ["a", "b"], "a": "1", "b": "2"}) == [
        "-11",
        "-12",
    ]
    assert var_expand("-$(c)$($(h))", [], {"h": ["a", "b"], "a": "1", "b": "2"}) == []


def test_edits():
    edits = var_edit_parse("L")
    assert edits.downshift
    assert not edits.upshift
    assert not edits.filemods

    edits = var_edit_parse("UP")
    assert not edits.downshift
    assert edits.upshift
    assert edits.parent
    assert edits.filemods

    edits = var_edit_parse("UQ")
    assert not edits.downshift
    assert edits.upshift
    assert edits.quote

    edits = var_edit_parse("QE=ab")
    assert edits.quote
    assert edits.empty == "ab"
    assert not edits.filemods

    edits = var_edit_parse(f"QE=ab{MC}J=ca")
    assert edits.quote
    assert edits.empty == "ab"
    assert edits.join == "ca"
    assert not edits.downshift

    edits = var_edit_parse(f"QE={MC}J=cad{MC}G=add")
    assert edits.quote
    assert edits.empty == ""
    assert edits.join == "cad"
    assert not edits.downshift

    edits = var_edit_parse(f"Q{MC}RG=acd{MC}M=baa{MC}U")
    assert edits.quote
    assert edits.f.root is None
    assert edits.f.grist == "acd"
    assert edits.f.member == "baa"
    assert edits.upshift
    assert edits.join is None
    assert edits.f.base == ""
    assert edits.filemods

    edits = var_edit_parse(f"QG=acd{MC}M=baa")
    assert edits.quote
    assert edits.f.root is None
    assert edits.f.grist == "acd"
    assert edits.f.member == "baa"
    assert edits.f.base is None
    assert edits.filemods

    edits = var_edit_parse(f"Q{MC}GB=bac{MC}S=sad{MC}D=dir{MC}U")
    assert edits.quote
    assert edits.upshift
    assert edits.filemods
    assert edits.join is None
    assert edits.f.grist is None
    assert edits.f.base == "bac"
    assert edits.f.suffix == "sad"
    assert edits.f.directory == "dir"

    edits = var_edit_parse(f"Q{MC}GB=bac{MC}S=sad{MC}D")
    assert edits.quote
    assert not edits.upshift
    assert edits.filemods
    assert edits.join is None
    assert edits.f.grist is None
    assert edits.f.base == "bac"
    assert edits.f.suffix == "sad"
    assert edits.f.directory is None


def test_paths():
    p = Pathname()
    orig = f"{S}one{S}two{S}three{S}file.c"
    p.parse(orig)
    assert p.build() == orig
    p.keep_only_parent()
    assert p.build() == f"{S}one{S}two{S}three"

    orig = f"<g1>one{S}two{S}three{S}file.c<mem>"
    p.parse(orig)
    assert p.build() == orig
    p.root = f"{S}add{S}"
    assert p.build() == f"<g1>{S}add{S}one{S}two{S}three{S}file.c<mem>"


def test_var_edits():
    assert var_expand("$(a:U)", [], {"a": ["ab"]}) == ["AB"]
    assert var_expand("$(a:E=cd)", [], {"a": []}) == ["cd"]
    assert var_expand("$(a:Q)", [], {"a": ["ab\\cd"]}) == ["ab\\\\cd"]
    assert var_expand("$(a:L)", [], {"a": ["AB", "CD"]}) == ["ab", "cd"]
    assert var_expand("$(a:L)e", [], {"a": ["AB", "CD"]}) == ["abe", "cde"]
    assert var_expand("y$(a:L)e", [], {"a": ["AB", "CD"]}) == ["yabe", "ycde"]
    assert var_expand("$(a:J=!) $(b)", [], {"a": ["AB", "CD"], "b": ["ab", "cd"]}) == [
        "AB!CD ab",
        "AB!CD cd",
    ]


def test_vms_paths_parent():
    p = Pathname(is_vms=True)
    orig = "src:[one]file.c"
    p.parse(orig)
    assert p.build() == orig
    p.keep_only_parent()
    assert p.build() == "src:[one]"

    p = Pathname(is_vms=True)
    orig = "src:[one]"
    p.parse(orig)
    assert p.build() == orig
    p.keep_only_parent()
    assert p.build() == "src:[000000]"

    p = Pathname(is_vms=True)
    orig = "src:[one.two]"
    p.parse(orig)
    assert p.build() == orig
    p.keep_only_parent()
    assert p.build() == "src:[one]"

    p = Pathname(is_vms=True)
    orig = "src:[000000]"
    p.parse(orig)
    assert p.build() == orig
    p.keep_only_parent()
    assert p.build() == "src:[000000]"

    p = Pathname(is_vms=True)
    orig = "<g1>dev:[one.two.three]file.c<mem>"
    p.parse(orig)
    assert p.build() == orig
    p.root = "[add:"
    assert p.build() == "<g1>dev:[one.two.three]file.c<mem>"


def test_vms_paths_root():
    p = Pathname(is_vms=True)
    orig = "[.one.two.three]file.c"
    p.parse(orig)
    assert p.build() == orig
    p.root = "src:[add]"
    assert p.build() == "src:[add.one.two.three]file.c"

    p = Pathname(is_vms=True)
    orig = "[one.two.three]file.c"
    p.parse(orig)
    assert p.build() == orig
    p.root = "src:"
    assert p.build() == "src:[one.two.three]file.c"

    p = Pathname(is_vms=True)
    orig = "one.c"
    p.parse(orig)
    assert p.build() == orig
    p.root = "src:"
    assert p.build() == "src:one.c"

    p = Pathname(is_vms=True)
    orig = "[.rel]"
    p.parse(orig)
    assert p.build() == orig
    p.root = "src:[one]"
    assert p.build() == "src:[one.rel]"

    p = Pathname(is_vms=True)
    orig = "[.rel]file.c;1"
    p.parse(orig)
    assert p.build() == orig
    p.root = "src:[one]"
    assert p.build() == "src:[one.rel]file.c;1"



import tempfile
import os
from os import sep as S

from jamp.classes import Target, State
from jamp.executors import run


def test_global_locate_search():
    state = State()
    target = Target.bind(state, "test.c")
    state.vars.set("LOCATE", ["/tmp"])
    assert target.search(state) == f"{S}tmp{S}test.c"

    state.vars.set("LOCATE", [])

    with tempfile.NamedTemporaryFile() as f:
        target_temp = Target.bind(state, os.path.basename(f.name))
        tempdir = os.path.dirname(f.name)
        state.vars.set("SEARCH", ["/proc", tempdir])
        assert target_temp.search(state) == f.name

    state.vars.set("SEARCH", [])
    assert target.search(state) == "test.c"


def test_target_locate_search():
    state = State()
    target = Target.bind(state, "test.c")
    target.vars["LOCATE"] = ["/tmp"]
    assert target.search(state) == f"{S}tmp{S}test.c"

    target.vars["LOCATE"] = []

    with tempfile.NamedTemporaryFile() as f:
        target_temp = Target.bind(state, os.path.basename(f.name))
        tempdir = os.path.dirname(f.name)
        target_temp.vars["SEARCH"] = ["/proc", tempdir]
        assert target_temp.search(state) == f.name

    target.vars["SEARCH"] = []
    assert target.search(state) == "test.c"


def test_linking_actions():
    rules = """
    rule OneCompile
    {
        rule Second
        {
            Echo $(1) ;
        }

        Clean clean : $(<) ;
        Compile1 $(<) : $(>) ;
        Compile2 $(<) : $(>) ;
    }

    rule Compile1
    {
        Echo "compile 1 called!" ;
    }

    actions Compile1 {
        echo "1"  $(>) $(<)
    }

    actions Compile2 {
        echo "2"  $(>) $(<)
    }

    actions OneCompile {
        echo "3" $(>) $(<) ;
    }

    OneCompile test.c : test.h ;
    Depends all : test.c ;
    Includes test.c : test.h ;
    """

    state = State()
    run(state, state.parse_and_compile(rules))
    print(state.targets)



import pytest


def pytest_configure(config):
    import sys

    sys._called_from_test = True


def pytest_unconfigure(config):
    import sys

    if hasattr(sys, '_called_from_test'):
        delattr(sys, '_called_from_test')


@pytest.fixture(autouse=True)
def run_around_tests():
    from jamp.jam_builtins import Builtins

    Builtins.clear_output()

    yield



import re
import os
import subprocess as sp
import pickle

from functools import cache
from typing import Optional

headers_cache = None
headers_cache_loaded = None
FN_CACHE = "jamp_saved_headers.cache"
HEADERS_CACHE_VERSION = 1


def load_headers_cache():
    global headers_cache
    global headers_cache_loaded

    if headers_cache is not None or headers_cache_loaded:
        return

    if os.path.exists(FN_CACHE):
        with open(FN_CACHE, "rb") as f:
            headers_cache = pickle.load(f)
            version = headers_cache.get("jamp_version")
            if version != HEADERS_CACHE_VERSION:
                headers_cache = {}
                headers_cache["jamp_version"] = HEADERS_CACHE_VERSION
                print(
                    f"jamp: headers cache was invalidated, new version is {HEADERS_CACHE_VERSION}"
                )
    else:
        headers_cache = {}

    headers_cache_loaded = True


def save_headers_cache():
    if headers_cache:
        try:
            with open(FN_CACHE, "wb") as f:
                pickle.dump(headers_cache, f)
        except Exception as e:
            print(f"jamp: could not save to headers cache to {FN_CACHE}: {e}")
            os.unlink(FN_CACHE)


def get_cached_headers(state, fn: str, timestamp: float):
    if headers_cache:
        data = headers_cache.get(fn)
        if data:
            saved_timestamp = data[0]
            if saved_timestamp < timestamp:
                if state.verbose:
                    print(
                        f"{fn} saved headers ignored, {saved_timestamp} < {timestamp}"
                    )

                # file was modified
                del headers_cache[fn]
                return None

            return data[1]


def target_find_headers(state, target, db: Optional[dict] = None) -> bool:
    from jamp.executors import exec_one_rule

    before_incs = len(target.includes)

    if target.boundname is None:
        return False

    if target.headers:
        return False

    hdrscan = state.vars.get("HDRSCAN", on_target=target)
    hdrrule = state.vars.get("HDRRULE", on_target=target)

    if not hdrscan or not hdrrule:
        return False

    lol = [[target.name]]
    headers = db.get(target.boundname) if db else None

    ts = None
    if headers is None:
        try:
            ts = os.path.getmtime(target.boundname)
        except FileNotFoundError:
            ts = None
        else:
            headers = get_cached_headers(state, target.boundname, ts)

    target.headers = headers or scan_headers(state, target.boundname, tuple(hdrscan))

    if state.debug_headers:
        if state.limit_target is not None:
            if state.limit_target in target.name:
                print(target.name, target.headers)
        else:
            print(target.name, target.headers)

    if target.headers:
        if ts is not None and headers is None:
            headers_cache[target.boundname] = (ts, target.headers)

        lol.append(target.headers)

        with target.overlay(state):
            for rule_name in hdrrule:
                exec_one_rule(state, rule_name, lol)

    if before_incs != len(target.includes):
        change_pairs = []

        for inc in target.includes:
            found_inc = inc.bind_location(state, strict=True)

            # Sometimes new found includes already have a some location but
            # just with different grist for example.
            # Then just use an old target.
            if found_inc is not None and inc != found_inc:
                change_pairs.append((inc, found_inc))

        if change_pairs:
            for old, new in change_pairs:
                target.includes.remove(old)
                target.includes.add(new)

        return True

    return False


def skip_include(state, boundname):
    sub_root = state.sub_root()
    if not boundname:
        return True

    if sub_root and not boundname.startswith(sub_root[0]):
        # skip outside headers scanning
        if state.verbose and boundname not in state.scan_skipped:
            if len(state.scan_skipped) == 0:
                print(
                    "info: headers outside the source root "
                    "directory will be skipped from headers scan"
                )
            print(f"skipped from headers scan: {boundname}")
            state.scan_skipped.add(boundname)

        return True

    return False


@cache
def scan_headers(state, fn: str, hdrscan: tuple):
    patterns = []
    for pattern in hdrscan:
        patterns.append(re.compile(pattern))

    if not os.path.exists(fn):
        if not state.verbose and not state.headers_complained:
            print(
                "jamp: errors while headers searching, "
                "use verbose option to turn on all messages"
            )
            state.headers_complained = True

        if state.verbose:
            print(f"jamp: {fn} not found while searching headers, skipped")

        return

    headers = []
    for pattern in patterns:
        with open(fn, errors="surrogateescape") as f:
            for i, line in enumerate(f):
                for m in re.finditer(pattern, line):
                    headers += list(m.groups())

    return headers


def scan_ripgrep_output(state, pattern):
    expect_fn = True
    headers = None
    fn = None
    skip_file = False

    res = {}

    lines = sp.check_output(["rg", "--heading", "-N", pattern])
    for line in lines.splitlines():
        if line == b"":
            expect_fn = True
            continue

        try:
            line = line.decode("utf8")
        except UnicodeDecodeError:
            continue

        if expect_fn:
            expect_fn = False
            fn = os.path.abspath(line)
            skip_file = fn.endswith(".yi")
            continue

        if skip_file:
            continue

        headers = res.setdefault(fn, [])

        for m in re.finditer(pattern, line):
            headers += list(m.groups())

    return res


def scan_grep_output(state, pattern):
    from jamp.paths import check_vms

    res = {}
    lines = sp.check_output(["grep", "-I", "-s", "-H", "-r", "-E", pattern])

    for line in lines.splitlines():
        try:
            line = line.decode("utf8")
        except UnicodeDecodeError:
            continue

        try:
            fn, match = line.split(":", 1)
        except ValueError:
            print(f"grep returned unexpected output: {line}")
            continue

        fn = os.path.abspath(fn) if not check_vms() else fn
        if fn.endswith(".yi"):
            continue

        headers = res.setdefault(fn, [])

        for m in re.finditer(pattern, match):
            headers += list(m.groups())

    return res



keywords = {
    "<": "LANGLE",
    "<=": "LANGLE_EQUALS",
    "=": "EQUALS",
    ">": "RANGLE",
    ">=": "RANGLE_EQUALS",
    "|": "BAR",
    "||": "BARBAR",
    ";": "SEMICOLON",
    ":": "COLON",
    "!": "BANG",
    "!=": "BANG_EQUALS",
    "?=": "QUESTION_EQUALS",
    "(": "LPAREN",
    ")": "RPAREN",
    "[": "LBRACKET",
    "]": "RBRACKET",
    "{": "LBRACE",
    "}": "RBRACE",
    "&": "AMPER",
    "&&": "AMPERAMPER",
    "+=": "PLUS_EQUALS",
    "actions": "ACTIONS",
    "bind": "BIND",
    "break": "BREAK",
    "case": "CASE",
    "continue": "CONTINUE",
    "default": "DEFAULT",
    "else": "ELSE",
    "existing": "EXISTING",
    "for": "FOR",
    "if": "IF",
    "ignore": "IGNORE",
    "in": "IN",
    "include": "INCLUDE",
    "local": "LOCAL",
    "maxline": "MAXLINE",
    "on": "ON",
    "piecemeal": "PIECEMEAL",
    "quietly": "QUIETLY",
    "return": "RETURN",
    "rule": "RULE",
    "switch": "SWITCH",
    "together": "TOGETHER",
    "updated": "UPDATED",
    "while": "WHILE",
}

SCAN_NORMAL = "n"
SCAN_STRING = "s"
SCAN_PUNCT = "p"
EOF = -1
EOL = "\n"


class LexerError(Exception):
    pass


class LexerToken(object):
    def __repr__(self):
        return f"LexToken({self.type},{self.value!r},{self.lineno},{self.lexpos})"


class Lexer:
    def __init__(self, filename=None):
        self.restart()
        self.lines = []
        self.filename = filename
        self.prevtoken = None

    def set_scanmode(self, mode):
        self.scanmode = mode

    def input(self, text: str):
        self.lines = text.split("\n")

    def nextline(self):
        self.prevpos = self.pos
        self.pos = 0
        self.prevlineno = self.lineno
        self.lineno += 1
        self.lexpos = 1

    def prev(self):
        self.pos = self.prevpos
        self.lineno = self.prevlineno

    def current_lineno(self):
        return self.lineno + 1

    def getchar(self):
        if self.lineno >= len(self.lines):
            return EOF

        line = self.lines[self.lineno]
        if self.pos >= len(line):
            self.nextline()
            return EOL

        self.prevpos = self.pos
        self.prevlineno = self.lineno
        self.pos += 1
        return line[self.prevpos]

    def get_string(self):
        # If scanning for a string (action's {}'s), look for the
        # closing brace.  We handle matching braces, if they match!
        nest = 1
        c = self.getchar()

        res = ""
        while c != EOF:
            if c == "{":
                nest += 1

            if c == "}":
                nest -= 1
                if nest == 0:
                    break

            res += c
            c = self.getchar()

        # We ate the ending brace -- regurgitate it
        if c != EOF:
            self.prev()

        if nest:
            raise LexerError("unmatched {} in action block")

        tok = LexerToken()
        tok.type = "STRING"
        tok.value = res

        return self.next_token(tok)

    def is_space(self, c):
        return c in set([" ", "\t", EOL, "\f", "\r"])

    def next_token(self, tok=None):
        if tok is None:
            self.finished = True
            return None

        tok.lexpos = self.lexpos
        tok.lexer = self
        tok.lineno = self.lineno + 1
        self.lexpos += 1

        if tok.type == "ACTIONS":
            self.actions_block = True

        if self.actions_block:
            if tok.type == "LBRACE":
                self.set_scanmode(SCAN_STRING)

            if tok.type == "STRING":
                self.set_scanmode(SCAN_NORMAL)
                self.actions_block = False

        self.prevtoken = tok
        return tok

    def restart(self):
        self.scanmode = SCAN_NORMAL

        self.pos = 0
        self.finished = False
        self.lineno = 0
        self.lexpos = 1
        self.prevpos = self.pos
        self.prevlineno = self.lineno
        self.actions_block = False

    def token(self):
        inquote = False

        if self.finished:
            return None

        if self.scanmode == SCAN_STRING:
            return self.get_string()

        c = self.getchar()

        while True:
            # Skip past white space
            while c != EOF and self.is_space(c):
                c = self.getchar()

            # Not a comment?  Swallow up comment line.
            if c != "#":
                break

            self.nextline()
            c = self.getchar()

        # c now points to the first character of a token
        if c == EOF:
            return self.next_token()

        # While scanning the word, disqualify it for (expensive)
        # keyword lookup when we can: $anything, "anything", \anything
        notkeyword = c == "$"

        # look for white space to delimit word */
        # "'s get stripped but preserve white space */
        # \ protects next character */
        res = ""
        while c != EOF and (inquote or not self.is_space(c)):
            if c == '"':
                # begin or end
                inquote = not inquote
                notkeyword = 1
            elif c != "\\":
                # normal char
                res += c
            else:
                c = self.getchar()
                if c == EOF:
                    break

                res += c
                notkeyword = True

            c = self.getchar()

        # Check obvious errors.
        if inquote:
            raise LexerError('unmatched " in string')

        # We looked ahead a character - back up.
        if c != EOF:
            self.prev()

        # scan token table
        # don't scan if it's obviously not a keyword or if its
        # an alphabetic when were looking for punctuation
        tok = LexerToken()
        tok.type = "ARG"
        tok.value = res

        if not notkeyword and not (res[0].isalpha() and self.scanmode == SCAN_PUNCT):
            if res in keywords:
                tok.type = keywords[res]

            if tok.type == "INCLUDE":
                if not (
                    self.prevtoken is None
                    or self.prevtoken.type in ("SEMICOLON", "LBRACE", "RBRACE")
                ):
                    tok.type = "ARG"

        return self.next_token(tok)



import subprocess as sp
import os
import tempfile
import shutil
import atexit

from contextlib import contextmanager
from jamp.build import main_cli


@contextmanager
def rel(path):
    curdir = os.getcwd()
    os.chdir(path)
    try:
        yield curdir
    finally:
        os.chdir(curdir)


def test_simple():
    d = "tests/test_simple"
    with rel(d):
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        output = sp.check_output("ninja")
        print(output)
        assert b"Two test.c" in output
        assert os.path.exists("test.c")
        output = sp.check_output("ninja")
        assert b"ninja: no work to do." in output
        output = sp.check_output(["ninja", "-t", "clean"])
        assert b"Cleaning... 1 files." in output


def test_subgen():
    d = "tests/test_subgen"
    with rel(d):
        os.environ["TOP"] = "."
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        output = sp.check_output("ninja")
        assert os.path.exists("app")
        output = sp.check_output("ninja")
        assert b"ninja: no work to do." in output


def test_dirs():
    d = "tests/test_dirs"
    with rel(d):
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        sp.check_output("ninja")
        assert os.path.exists("sub1/two.c")
        assert os.path.exists("sub2/three.c")
        sp.run(["ninja", "-t", "clean"])
        assert not os.path.exists("sub1/two.c")
        assert not os.path.exists("sub2/three.c")
        assert os.path.exists("sub1")


def test_copy_files():
    d = "tests/test_copy_files"
    with rel(d):
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        sp.check_output("ninja")
        assert os.path.exists("foo.so")
        sp.run(["ninja", "-t", "clean"])
        assert not os.path.exists("foo.so")


def test_multiline():
    d = "tests/test_multiline"
    with rel(d):
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        sp.check_output("ninja")
        assert os.path.exists("out.txt")
        sp.run(["ninja", "-t", "clean"])
        assert not os.path.exists("out.txt")


def test_simple_app():
    d = "tests/test_simple_app"
    with rel(d):
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        sp.check_output("ninja")
        assert os.path.exists("app")
        assert os.path.exists("libprint.a")
        assert os.path.exists("libsay.a")
        sp.run(["ninja", "-t", "clean"])
        assert not os.path.exists("app")


def test_math_example():
    d = "tests/test_math_example"
    with rel(d):
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        sp.check_output("ninja")
        assert os.path.exists("app")
        assert os.path.exists("libprint.a")
        sp.run(["ninja", "-t", "clean"])
        assert not os.path.exists("app")
        assert not os.path.exists("libprint.a")


def test_circular_inc():
    d = "tests/test_circular_inc"
    with rel(d):
        main_cli(skip_args=True)
        sp.run(["ninja", "-t", "clean"])
        sp.check_output("ninja")
        assert os.path.exists("app")
        sp.run(["ninja", "-t", "clean"])
        assert not os.path.exists("app")


def test_jam_compat():
    try:
        sp.check_output(["jam", "-v"])
    except:
        return

    files = ("tests/data/jam1.jam",)
    outputs = (
        b"""[1/1] One1 a.txt
a.txt b.txt
a.txt b.txt
a.txt b.txt
b.txt a.txt
""",
    )

    for i, f in enumerate(files):
        tdir = tempfile.mkdtemp(prefix="jamp")
        atexit.register(lambda: shutil.rmtree(tdir, ignore_errors=True))

        with rel(tdir) as old:
            out = os.path.join(tdir, "Jamfile")
            with open(os.path.join(old, f)) as fin:
                content = fin.read()
                with open(out, "w") as fout:
                    fout.write(content)

            main_cli(skip_args=True)
            ninja_output = sp.check_output("ninja")
            assert ninja_output == outputs[i]

            failed = False
            try:
                out = sp.check_output(["jam"])
            except:
                failed = True

            assert not failed, out



import argparse
import os
import sys
import subprocess as sp

from collections import OrderedDict

from jamp import executors, headers
from jamp.classes import State, Target, UpdatingAction
from jamp.paths import check_vms, escape_path, add_paths, check_windows


def parse_args(skip_args=False):
    parser = argparse.ArgumentParser(
        prog="jamp",
        description="Jam Build System (Python version)",
    )
    parser.add_argument("-b", "--build", action="store_true", help="call ninja")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("--profile", action="store_true", help="profile the execution")
    parser.add_argument(
        "--depfiles",
        action="store_true",
        help="use depfile feature of ninja (only Unix)",
    )
    parser.add_argument(
        "--no-headers-cache", action="store_true", help="do not cache found headers"
    )
    parser.add_argument(
        "-s",
        "--search-type",
        default="base",
        choices=["base", "ripgrep", "grep", "none"],
        help="headers search type (default is basic jam algorithm)",
    )
    parser.add_argument(
        "-d",
        "--debug",
        default=[],
        choices=["headers", "depends", "include", "env"],
        help="show headers",
        nargs="+",
    )
    parser.add_argument(
        "-t", "--target", default=None, help="limit target for debug info"
    )
    parser.add_argument(
        "--unwrap-phony",
        default=[],
        help=(
            "unwrap specified phony targets in deps (useful for debug, "
            "to find exact triggering input)"
        ),
        nargs="+",
    )
    parser.add_argument(
        "-f", "--jamfile", default="Jamfile", help="--specify jam file name"
    )
    parser.add_argument(
        "-e", "--env", action="append", help="--specify extra env variables"
    )
    args = parser.parse_args(args=[] if skip_args else None)
    return args


def main_app(args):
    """Main entrypoint"""

    curdir = os.path.abspath(os.getcwd())
    basedir = os.path.dirname(__file__)
    jambase = os.path.join(basedir, "Jambase")

    state = State(
        verbose=args.verbose,
        debug_headers="headers" in args.debug,
        debug_deps="depends" in args.debug,
        debug_include="include" in args.debug,
        debug_env="env" in args.debug,
        target=args.target,
        unwrap_phony=args.unwrap_phony,
    )
    jamfile = args.jamfile

    state.vars.set("JAMFILE", [jamfile])
    state.vars.set("JAMP_PYTHON", [sys.executable])
    state.vars.set("JAMP_OPTIONS", sys.argv[1:])
    state.vars.set("NINJA_ROOTDIR", [curdir])

    if args.depfiles:
        state.vars.set("ENABLE_DEPFILES", ["1"])

    for var in args.env or ():
        parts = var.split("=")
        state.vars.set(parts[0], [parts[1]])

    if not os.path.exists(jamfile):
        print("Jamfile not found")
        exit(1)

    with open(jambase) as f:
        jambase_contents = f.read()

    if args.verbose:
        print("...parsing jam files...")

    cmds = state.parse_and_compile(jambase_contents)

    if args.verbose:
        print("...execution...")

    executors.run(state, cmds)
    if args.verbose:
        print("...binding targets and searching headers...")

    if not args.no_headers_cache:
        headers.load_headers_cache()

    executors.bind_targets(state, search_headers=args.search_type)

    if not args.no_headers_cache:
        headers.save_headers_cache()

    all_target = Target.bind(state, "all")
    all_target.search_for_cycles(verbose=args.verbose)

    state.finish_steps()

    print(f"...found {len(state.targets)} target(s)...")
    if args.verbose:
        print("...writing build.ninja...")

    with open("build.ninja", "w") as f:
        ninja_build(state, f)

    if args.build:
        sp.run(["ninja"])


def ninja_build(state: State, output):
    """Write ninja.build"""

    from jamp.ninja_syntax import Writer

    writer = Writer(output, width=120)

    target: Target = None

    counter = 0
    commands_cache = {}

    for step in state.build_steps:
        upd_action: UpdatingAction = step[1]
        upd_action.name = f"{upd_action.action.name}{counter}".replace("+", "_")
        counter += 1

        full_cmd, oneliner = upd_action.get_command(state)

        # an optimization for simple rules with one command
        # group similar rules to one
        if upd_action.is_alone():
            found = False
            key = upd_action.action.name

            if key in commands_cache:
                saved = commands_cache[key]

                for name, cached_cmd in saved:
                    if full_cmd == cached_cmd:
                        # no need to create a new rule, we have similar
                        upd_action.name = name
                        found = True
                        break

                if found:
                    continue

            else:
                saved = commands_cache.setdefault(key, [])

            saved.append((upd_action.name, full_cmd))

        if not oneliner and check_windows():
            resp_fn = f"{upd_action.name}$step.bat"

            writer.rule(
                upd_action.name,
                command=f"cmd /C {resp_fn}",
                description=upd_action.description(),
                rspfile=resp_fn,
                rspfile_content=full_cmd,
                restat=upd_action.restat,
                generator=upd_action.generator,
            )
        elif not oneliner and check_vms():
            # rule can be reused from saved, need the unique number for the resp file name
            resp_fn = f"{upd_action.name}$step.com"

            writer.rule(
                upd_action.name,
                command=f"@{resp_fn}",
                description=upd_action.description(),
                rspfile=resp_fn,
                rspfile_content=full_cmd,
                restat=upd_action.restat,
                generator=upd_action.generator,
            )
        else:
            # set depfile if needed
            for t in upd_action.targets:
                depfile = t.vars.get("DEPFILE")
                if depfile:
                    upd_action.depfile = depfile
                    break

            writer.rule(
                upd_action.name,
                full_cmd,
                restat=upd_action.restat,
                generator=upd_action.generator,
                depfile=upd_action.depfile,
                description=upd_action.description(),
            )

    phonies = {}
    gen_headers = {}

    for dep in state.targets["_gen_headers"].depends:
        if dep.boundname:
            gen_headers[dep.boundname] = None

    for target in state.targets.values():
        implicit, order_only = (escape_path(i) for i in target.get_dependency_list(state))
        if target.notfile:
            kwargs = {}
            if target.is_dirs_target:
                kwargs["order_only"] = (implicit | order_only)
            else:
                kwargs["order_only"] = order_only
                kwargs["implicit"] = implicit

            writer.build(target.name, "phony", **kwargs)
            phonies[target.name] = True

    for target in state.targets.values():
        if target.collection is not None:
            if target.collection_name() in phonies:
                continue

            implicit_deps = (escape_path(i) for i in target.collection[0])
            order_only_deps = (escape_path(i) for i in target.collection[1])
            writer.build(target.collection_name(), "phony", implicit=implicit_deps,
                         order_only=order_only_deps)
            phonies[target.collection_name()] = True

    for stepnum, step in enumerate(state.build_steps):
        outputs = OrderedDict()
        targets, upd_action = step

        for target in targets:
            if not target.boundname:
                continue

            outputs[target.boundname] = None

        if len(outputs) == 0:
            continue

        all_implicit = set()
        all_order_only = set()

        for target in targets:
            implicit, order_only = target.get_dependency_list(state, outputs=outputs)
            add_paths(all_implicit, implicit)
            add_paths(all_order_only, order_only)

        inputs = OrderedDict()

        for source in upd_action.sources:
            inputs[escape_path(source.boundname or source.name)] = None

        res_implicit = set()
        res_order_only = set()

        for dep in all_implicit:
            if dep in inputs:
                continue

            if dep in gen_headers:
                res_order_only.add(dep)
            else:
                res_implicit.add(dep)

        for dep in all_order_only:
            if dep in inputs:
                continue

            res_order_only.add(dep)

        variables = None

        if check_vms() or check_windows():
            variables = {"step": stepnum}

        writer.build(
            (escape_path(i) for i in outputs.keys()),
            upd_action.name,
            inputs.keys(),
            implicit=res_implicit,
            order_only=res_order_only,
            variables=variables,
        )

    writer.default("all")


def main_cli(skip_args=False):
    """Command line entrypoint"""

    args = parse_args(skip_args=skip_args)
    if args.profile:
        import cProfile

        ctx = {"args": args, "main_app": main_app}
        cProfile.runctx("main_app(args)", ctx, {})
    else:
        try:
            main_app(args)
        except KeyboardInterrupt:
            print("jamp: interrupted")



import os
from typing import List, Union, Set
from functools import cache

from jamp.paths import Pathname, check_vms, check_windows
from jamp.headers import target_find_headers, skip_include

PATH_VARS = set(
    ["PATH", "LD_LIBRARY_PATH", "PKG_CONFIG_PATH", "CLASSPATH", "PYTHONPATH"]
)


def is_subdir(path: str, potential_subdir: str):
    # Normalize paths to handle different path structures
    norm_path = os.path.normpath(path).replace(os.sep, "/")
    norm_subdir = os.path.normpath(potential_subdir).replace(os.sep, "/")
    return norm_subdir.startswith(norm_path)


def remove_overlapping(dirs: List):
    # Sort dirs by length in descending order
    dirs_sorted = sorted(dirs, key=len, reverse=True)

    result = []

    for dry in dirs_sorted:
        # Check if dir is not a subdirectory of any existing directory in result
        if not any(is_subdir(d, dry) for d in result):
            result.append(dry)

    return result


class State:
    def __init__(
        self,
        verbose=False,
        debug_headers=False,
        debug_deps=False,
        debug_include=False,
        debug_env=False,
        target=None,
        unwrap_phony=None,
    ):
        self.headers_complained = False
        self.verbose = verbose
        self.vars = Vars(debug_env=debug_env)
        self.rules = {}
        self.actions = {}
        self.targets = {}
        self.current_rule = None
        self.params = None
        self.always_build = set()
        self.build_steps = []
        self.debug_headers = debug_headers
        self.debug_deps = debug_deps
        self.debug_include = debug_include
        self.limit_target = target
        self.unwrap_phony = unwrap_phony

        # reverse location->target map
        self.target_locations = {}

        # skipped from scanning headers, just a cache
        self.scan_skipped = set()

    @cache
    def sub_root(self):
        sub_root = self.vars.get("SUBDIR_ROOT")
        if not sub_root:
            sub_root = self.vars.get("NINJA_ROOTDIR")

        if self.verbose:
            print(f"source root: {sub_root}")

        return sub_root

    def parse_and_compile(self, contents: str, filename=None):
        from jamp.jam_syntax import parse
        from jamp.compile import compile

        ast = parse(contents, filename=filename)
        cmds = compile(self, ast)
        return cmds

    def get_target(self, name):
        return self.targets.get(name)

    def add_action_for_target(
        self, target, action_name, generator=False, sources=None
    ):
        sources = sources or []
        action = self.actions[action_name]
        upd_action = UpdatingAction(action, sources)
        upd_action.targets = [target]
        upd_action.generator = generator

        if target.build_step is None:
            step = ([target], upd_action)
            self.build_steps.append(step)
            target.build_step = step
        else:
            prev_upd_action = target.build_step[1]
            prev_upd_action.link(upd_action)

        return upd_action

    def finish_steps(self):
        target = Target.bind(self, "dirs")

        if target.collected_dirs:
            dirs = remove_overlapping(target.collected_dirs)

            for dry in dirs:
                dir_target = Target.bind(self, dry)
                dir_target.boundname = dry
                dir_target.is_dir = True

                action = self.add_action_for_target(
                    dir_target,
                    "MkDirWhenNotExists",
                    generator=True,
                )
                action.targets = [target]
                target.depends.add(dir_target)


class Vars:
    delete_vars = ["LS_COLORS", "GITHUB_TOKEN"]

    def __init__(self, debug_env=False):
        self.debug_env = debug_env
        self.scopes = []
        self.scope = {}
        self.global_scope = self.scope
        self.set_basic_vars()

        # setting current targets will force to using target variables
        self.current_context = []

    def split_path(self, val):
        return val.split(os.path.pathsep)

    def set_basic_vars(self):
        import os
        import platform

        if check_windows():
            import nt

            self.scope.update(nt.environ.copy())
        else:
            self.scope.update(os.environ.copy())

        for v in self.delete_vars:
            if v in self.scope:
                del self.scope[v]

        match platform.system():
            case "Linux" | "Solaris" | "AIX" | "Darwin":
                self.scope["UNIX"] = "1"
            case "OpenVMS":
                self.scope["VMS"] = "1"
            case "Windows":
                self.scope["NT"] = "1"

        self.scope["OSPLAT"] = platform.machine()
        self.scope["OS"] = platform.system().upper()
        self.scope["JAMUNAME"] = platform.uname()
        self.scope["JAMVERSION"] = "2.6.1"
        self.scope["JAMCOUNTER"] = "<NINJA_SIGIL>step"

        for k, v in self.scope.items():
            if k in PATH_VARS:
                self.scope[k] = self.split_path(v)

        if self.debug_env:
            for key, val in self.scope.items():
                print(f"{key}={val}")

    def __repr__(self):
        return f"current scope: {self.scope}\nscopes: {self.scopes}"

    def set(self, name: str, value: str | None):
        if not isinstance(name, str):
            raise Exception(f"vars_set: expected str value for key name: got {name}")

        if not isinstance(value, list):
            raise Exception("vars_set: expected list for value")

        if isinstance(value, list) and len(value) and isinstance(value[0], list):
            raise Exception(f"can't store LOL as value for {name}: got {value}")

        if name in self.scope:
            # something local
            self.scope[name] = value
        else:
            # check in upper levels
            for level in reversed(self.scopes):
                if name in level:
                    level[name] = value
                    return

            # not defined, goes to global
            self.global_scope[name] = value

    def get_scope(self, name: str):
        if not isinstance(name, str):
            raise Exception(
                f"vars_get_scope: expected str value for key name: got {name}"
            )

        res_scope = None

        if name in self.scope:
            # something local
            res_scope = self.scope
        else:
            # check in upper levels
            for level in reversed(self.scopes):
                if name in level:
                    res_scope = level
                    break

            # not defined, check global
            if res_scope is None and name in self.global_scope:
                return self.global_scope

        # probably value in symbols, but not read, so read it and set in the current scope
        if res_scope is None:
            value = self.check_vms_symbol(name)
            if value is not None:
                res_scope = self.global_scope

        return res_scope

    def check_vms_symbol(self, name):
        if check_vms():
            import vms.lib

            status, val = vms.lib.get_symbol(name)
            if status == 1:
                self.global_scope[name] = [val]
                if self.debug_env:
                    print(f"{name}={val}")

                return [val]

    def set_local(self, name: str, value: str | None):
        value = value if value is not None else []
        self.scope[name] = value

    def get(self, name: str, on_target=None):
        if not isinstance(name, str):
            raise Exception(f"vars_get: expected str value for key name: got {name}")

        if on_target:
            self.current_context.append(on_target.vars)

        res = None
        if self.current_context:
            for ctx in reversed(self.current_context):
                if name in ctx:
                    res = ctx.get(name)
                    break

        if on_target:
            self.current_context.pop()

        if res is None:
            if name in self.scope:
                res = self.scope.get(name)
            else:
                for level in reversed(self.scopes):
                    if name in level:
                        res = level.get(name)
                        break

        if res is None:
            val = self.check_vms_symbol(name)
            if val is not None:
                return val

        return res if res else []

    def push(self):
        self.scopes.append(self.scope)
        self.scope = {}

    def pop(self):
        self.scope = self.scopes.pop()


class Rule:
    def __init__(self, name: str, params, commands):
        self.name = name
        self.params = params
        self.commands = commands

    def execute(self, state: State):
        from jamp.executors import run

        run(state, self.commands)

    def __repr__(self):
        return f"Rule {self.name}"


class Actions:
    def __init__(self, name: str, flags=None, bindlist=None, commands=None):
        self.name = name
        self.flags = flags
        self.bindlist = bindlist
        self.commands = commands

    def __repr__(self):
        return f"Actions {self.name}"


class Exec:
    """Just a wrapper for function and arguments"""

    def __init__(self, func, args):
        self.func = func
        self.args = args

    def execute(self, state):
        return self.func(state, *self.args)

    def __repr__(self):
        return f"F:{self.func.__name__}"


class UnderTarget:
    def __init__(self, state: State, target):
        self.state = state
        self.target = target

    def __enter__(self):
        self.state.vars.current_context.append(self.target.vars)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.state.vars.current_context.pop()
        return True


class Target:
    existing_paths = {}

    @classmethod
    def bind(cls, state: State, name: str, notfile=False):
        if name in state.targets:
            return state.targets[name]

        target = Target(name, notfile=notfile)
        state.targets[name] = target
        return target

    def overlay(self, state: State):
        return UnderTarget(state, self)

    def is_order_only(self):
        return self.is_output and self.is_header

    def check_if_dir(self):
        if self.is_dir:
            return True

        if self.boundname and check_vms() and self.boundname.endswith("]"):
            return True

        return False

    def __init__(self, name: str, notfile=False):
        self.name: str = name
        self.depends: Set[Target] = set()
        self.includes: Set[Target] = set()
        self.boundname: Union[None, str] = None
        self.updating_actions: List[UpdatingAction] = []
        self.build_step: tuple = None

        # Created my MkDir rule
        self.is_dir = False

        # The suffix was checked and set to True if it's header
        self.is_header = False

        # This target used somewhere as an output
        self.is_output = False

        # True if this is the main 'dirs' target
        self.is_dirs_target = False

        # for mkdir
        self.collected_dirs = None

        # Force generator option to ninja
        self.generated = False

        # Force restat option to ninja
        self.restat = False

        # collection is optimization, if this target is include and it depends on other
        # files just create a phony target with this target and all others
        self.collection = None

        # Dependencies cache after get_dependency_list call without outputs
        self.deps = None

        # Target level variables (ON <target> calls and etc)
        self.vars = {}

        # Temporary rule called on this target
        self.temporary = False

        # NotFile rule called on this target
        self.notfile = notfile  # phony

        # Found headers
        self.headers = None

        # Circular search
        self.circular_visited = False

        # NoCare rule
        self.nocare = False

        # NoUpdate rule
        self.noupdate = False

    def collection_name(self):
        t = self.name

        if check_vms() or check_windows():
            # : is a special escape for VMS paths
            # for Windows - disk drives (C:)
            t = self.name.replace(":", "_").lower()

        return f"_{t}_"

    def not_searchable(self):
        return "SEARCH" not in self.vars and "LOCATE" not in self.vars

    def get_dependency_list(self, state: State, level=0, outputs=None):
        """Ninja level dependency list"""

        implicit, order_only = set(), set()
        use_cached = outputs is None or len(outputs) == 1

        if level == 10:
            # do not go too deep for includes
            return (implicit, order_only)

        if use_cached and self.deps:
            return self.deps

        for t in self.depends:
            depval = None

            if t.notfile:
                if state.unwrap_phony and t.name in state.unwrap_phony:
                    phony_deps_impl, phony_deps_order = t.get_dependency_list(state)
                    implicit |= phony_deps_impl
                    order_only |= phony_deps_order
                else:
                    depval = t.name
            elif t.nocare and t.not_searchable():
                continue
            elif t.boundname:
                if not self.is_dirs_target and t.check_if_dir():
                    implicit.add("dirs")
                    continue

                depval = t.boundname
            elif t.nocare:
                continue

            if depval:
                if outputs is not None and depval in outputs:
                    continue
                elif t.noupdate:
                    order_only.add(depval)
                else:
                    implicit.add(depval)

        if not self.notfile:
            for t in self.includes:
                if use_cached and t.collection is not None:
                    implicit.add(t.collection_name())
                    continue

                depval = None
                if t.notfile:
                    depval = t.name
                elif t.boundname and t.boundname in state.target_locations:
                    depval = t.boundname
                # elif t.boundname and os.path.isfile(t.boundname):
                #    depval = t.boundname
                elif t.nocare:
                    continue

                if depval is None:
                    continue

                if outputs and depval in outputs:
                    continue

                if len(t.depends) or len(t.includes):
                    inner_deps_impl, inner_deps_order = t.get_dependency_list(
                        state, level=level + 1, outputs=outputs
                    )

                    if not use_cached:
                        implicit |= inner_deps_impl
                        order_only |= inner_deps_order
                    elif len(inner_deps_impl) or len(inner_deps_order):
                        t.collection = (
                            set((depval,)) | inner_deps_impl,
                            set(inner_deps_order),
                        )
                        depval = t.collection_name()

                if t.noupdate:
                    order_only.add(depval)
                else:
                    implicit.add(depval)

            # collect dependencies from sources which are not built
            for dep in self.depends:
                if dep.notfile:
                    continue
                elif dep.build_step is None:
                    built_deps_impl, built_deps_order = dep.get_dependency_list(
                        state, outputs=outputs
                    )
                    implicit |= built_deps_impl
                    order_only |= built_deps_order

        if state.debug_deps:
            if state.limit_target is not None:
                if state.limit_target in self.name:
                    print(self.name, implicit, order_only)
            else:
                print(self.name, implicit, order_only)

        if not use_cached:
            return (implicit, order_only)

        self.deps = (frozenset(implicit), frozenset(order_only))
        return self.deps

    def find_headers(self, state: State, level=0, db=None):
        if level == 10:
            # do not go too deep in searching
            return

        if self.is_output:
            # do not scan output targets
            # if file was already built, then we scan it, and remove that file
            # ninja could fail
            return

        if self.headers is not None:
            # do not search more than one time no each target
            return

        self.headers = []
        found = target_find_headers(state, self, db=db)

        if found:
            for inc in self.includes:
                if skip_include(state, inc.boundname):
                    continue

                inc.find_headers(state, level=level + 1, db=db)

    def bind_location(self, state: State, strict=False):
        """Returns a target if was found the a same location"""

        if not self.boundname:
            self.boundname = self.search(state, strict=strict)

        if self.boundname:
            if self.boundname in state.target_locations:
                return state.target_locations[self.boundname]

            state.target_locations[self.boundname] = self

        if self.is_order_only():
            gen_headers = state.targets.get("_gen_headers")
            if gen_headers:
                gen_headers.depends.add(self)

    def search(self, state: State, strict=False):
        """
        Using SEARCH and TARGET variables on the target try to construct
        the full name.
        Or just return the name of the target if strict is not True
        'strict' argument is used for headers when we need something more correct than
        just a name.
        """

        if self.notfile:
            return None

        if not self.name:
            return None

        path = Pathname()
        path.parse(self.name)

        # remember if it's header
        self.is_header = path.suffix in (".h", ".hpp", ".hh")

        if path.member:
            return None

        # remove the grist part in the filename
        path.grist = ""

        locate = state.vars.get("LOCATE", on_target=self)

        if locate:
            locate_dir = locate[0]
            path.root = locate_dir
            res_path = path.build(binding=True)

            return res_path
        else:
            search = state.vars.get("SEARCH", on_target=self)
            for search_dir in search:
                locate_dir = search_dir
                path.root = locate_dir
                res_path = path.build(binding=True)

                if res_path in state.target_locations:
                    # this could be a generated file, and if it's in targets just return that path
                    return res_path
                elif os.path.exists(res_path):
                    return res_path

        # recreate
        path = Pathname()
        path.parse(self.name)
        path.grist = ""
        res_path = path.build(binding=True)

        if strict and os.path.exists(res_path):
            return res_path

        return res_path if not strict else None

    def add_depends(self, state: State, targets: list):
        for target in targets:
            if isinstance(target, str):
                target = Target.bind(state, target)

            if target == self:
                continue

            self.depends.add(target)

    def add_includes(self, state: State, targets: list):
        for target in targets:
            if isinstance(target, str):
                target = Target.bind(state, target)

            if target == self:
                continue

            self.includes.add(target)

    def search_for_cycles(self, verbose=False, graph=None):
        try:
            import networkx
        except ImportError:
            print("!!! install `networkx` package to fix circular dependency errors")
            return

        if self.circular_visited:
            return

        self.circular_visited = True

        top = False
        if graph is None:
            top = True
            graph = networkx.DiGraph()

        for inc in self.includes:
            graph.add_edge(self, inc)
            inc.search_for_cycles(graph=graph)

        for dep in self.depends:
            dep.search_for_cycles(graph=graph)

        if top:
            for cycle in networkx.simple_cycles(graph):
                if verbose:
                    print(f"removed circular dependency: {cycle[0]} from {cycle[-1]}")

                try:
                    cycle[-1].includes.remove(cycle[0])
                except KeyError:
                    pass

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return f"T:{self.name}"


class UpdatingAction:
    windows_cmd_join = "$\n$^"

    def __init__(self, action: Actions, sources: list):
        self.action = action
        self.sources = sources
        self.base = None
        self.next: List[UpdatingAction] = []
        self.targets = []
        self.command = None
        self.restat = False
        self.generator = False
        self.depfile = None

        # for actions .. bind
        self.bindvars = None
        self.bindparams = None

    def link(self, upd_action):
        self.next.append(upd_action)
        upd_action.base = self

    def bound_params(self):
        res = []
        if self.targets:
            res.append([target.boundname for target in self.targets])
        else:
            res.append([])

        res.append([source.boundname for source in self.sources if source.boundname])
        return res

    def description(self):
        names = set([self.action.name])
        for n in self.next:
            names.add(n.action.name)

        return " & ".join(names) + " $out"

    def modify_vms_paths(self, state):
        """If path doesn't have directory, make it current"""

        if not self.bindparams:
            return

        for var, value in self.bindparams.items():
            modified = []
            for item in value:
                has_dir = ":" in item or "[" in item
                if has_dir:
                    modified.append(item)
                else:
                    modified.append("[]" + item)

            self.bindparams[var] = modified

    def is_alone(self):
        return not (bool(self.next) or bool(self.base))

    def prepare_lines(self, state, comment_sym="#"):
        from jamp.expand import var_string

        saved_context = state.vars.current_context
        state.vars.current_context = []
        for t in self.targets:
            state.vars.current_context.append(t.vars)

        if self.bindparams:
            state.vars.current_context.append(self.bindparams)

        lines = self.action.commands

        for line in lines.split("\n"):
            line = line.strip()
            if not line:
                continue
            if line.startswith(comment_sym):
                continue

            line = var_string(
                line,
                self.bound_params(),
                state.vars,
                alone=self.is_alone(),
            )
            line = line.replace("$", "$$")
            line = line.replace("<NINJA_SIGIL>", "$")

            if not line:
                continue

            yield line

        state.vars.current_context = saved_context

    def prepare_action(self, state: State):
        quotes = []
        concat = ""

        start_new_command = False
        for line in self.prepare_lines(state):
            if start_new_command:
                concat += " ; $\n "

            start_new_command = False

            # watch for open quotes
            for c in line:
                if c in ["'", '"', "`"]:
                    if quotes and quotes[-1] == c:
                        quotes.pop()
                    else:
                        # a new quote started
                        quotes.append(c)

            if line.endswith("\\"):
                concat += line[:-1]
            elif line.endswith("&&"):
                concat += line + " "
            elif line.endswith(";"):
                concat += line + " "
            elif line.endswith("("):
                concat += line + " "
            elif line.endswith("|"):
                concat += line + " "
            elif line.endswith("{"):
                concat += line + " "
            elif line == "then" or line.endswith(" then"):
                concat += line + " "
            elif line == "do" or line.endswith(" do"):
                concat += line + " "
            elif line == "else" or line.endswith(" else"):
                concat += line + " "
            elif len(quotes):
                concat += line + " "
            else:
                concat += line
                start_new_command = True

        return concat, True

    def prepare_windows_action(self, state: State):
        """not tested"""
        quotes = []
        concat = ""

        new_cmd = None

        for line in self.prepare_lines(state, comment_sym="REM"):
            if new_cmd is not None:
                concat += new_cmd

            new_cmd = None

            # watch for open quotes and redirections
            for c in line:
                if c in ["'", '"']:
                    if quotes and quotes[-1] == c:
                        quotes.pop()
                    else:
                        # a new quote started
                        quotes.append(c)

            if line.endswith("^"):
                concat += line[:-1]
            elif len(quotes):
                concat += line
            else:
                concat += line
                new_cmd = self.windows_cmd_join

        return concat, False

    def prepare_vms_action(self, state: State):
        quotes = []
        concat = "$$ "

        oneliner = True
        add_newline = False
        for line in self.prepare_lines(state, comment_sym="!"):
            # watch for open quotes
            if add_newline:
                oneliner = False
                concat += " $\n$^$$"

            add_newline = False
            for c in line:
                if c in ['"']:
                    if quotes and quotes[-1] == c:
                        quotes.pop()
                    else:
                        # a new quote started
                        quotes.append(c)

            if line.endswith("-"):
                concat += line[:-1]
            elif len(quotes):
                concat += line + " "
            else:
                # $^ is a hack to samurai (github.com/ildus/samurai)
                # which adds a proper newline in a script
                concat += line
                add_newline = True

        return concat, oneliner

    def process_bind_vars(self, state):
        """Set actual boundnames for BIND params in actions"""

        if self.bindparams:
            return

        if not self.bindvars:
            return

        bindparams = {}

        for target in self.targets:
            for var in self.bindvars:
                bind_values = state.vars.get(var, on_target=target)
                if not bind_values:
                    continue

                res = []
                for val in bind_values:
                    bindtarget = Target.bind(state, val)
                    if bindtarget.boundname:
                        bindtarget.bind_location(state)

                    # if bind was successfull
                    if bindtarget.boundname:
                        res.append(bindtarget.boundname)

                if res:
                    bindparams[var] = res

        if bindparams:
            self.bindparams = bindparams

            if check_vms():
                self.modify_vms_paths(state)

    def get_command(self, state: State, force_vms=False, force_windows=False):
        self.process_bind_vars(state)

        still_oneliner = True

        if not self.command:
            if force_vms or check_vms():
                base_lines, oneliner = self.prepare_vms_action(state)
            elif force_windows or check_windows():
                base_lines, oneliner = self.prepare_windows_action(state)
            else:
                base_lines, oneliner = self.prepare_action(state)

            if not oneliner:
                still_oneliner = False

            if self.next:
                still_oneliner = False

                for next_upd_action in self.next:
                    lines, _ = next_upd_action.get_command(state)
                    if check_vms():
                        base_lines += "$\n$^" + lines
                    elif check_windows():
                        base_lines += self.windows_cmd_join + lines
                    else:
                        base_lines += " ; $\n" + lines

            if check_vms():
                # just an empty line at the end
                base_lines += "$\n$^$$"

            self.command = base_lines

        return self.command, still_oneliner
